"""
Block filling algorithm. 
Date: 2023-02-05
Authors: Mauricio Pereira, Antonio Maria D'Altri
Tested with Python 3.+. All packages should be readily available using Conda packages.
"""

import pandas as pd
import numpy as np
from scipy import signal
import datetime, os

from matplotlib.axes._axes import _log as matplotlib_axes_logger
matplotlib_axes_logger.setLevel('ERROR')

import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import cm

import warnings
warnings.filterwarnings('ignore')

import networkx as nx

plt.rcParams["text.usetex"] = False
plt.rcParams["font.size"] = 8
lw = 1
ms = 0.5
dpi = 300

imgw = 6
imgh = 3

masonry_colors = np.array([
[141/255.0, 82/255.0, 54/255.0],
[144/255.0, 91/255.0, 63/255.0],
[181/255.0, 122/255.0, 97/255.0],
[185/255.0, 166/255.0, 136/255.0],
[153/255.0, 109/255.0, 81/255.0],
[90/255.0, 61/255.0, 52/255.0],
[77/255.0, 71/255.0, 66/255.0],
[162/255.0, 140/255.0, 118/255.0],
[179/255.0, 103/255.0, 71/255.0],
[204/255.0, 141/255.0, 97/255.0],
[204/255.0, 141/255.0, 97/255.0]])

masonry = mpl.colors.LinearSegmentedColormap.from_list('masonry', masonry_colors, N=masonry_colors.shape[0], gamma=1.0)


def probs(sample_wall, block_sizes):
    """
    Get sampling probabilities of each block type as A_block/A_total.
    Inputs:
        sample_wall: H x W array
        block_sizes: Nblocks x 3 array
    Outputs:
        p: Nblocks array
    """

    Nblock_types = block_sizes.shape[0]
    p = []
    for r in range(Nblock_types):
        block_area = block_sizes[r,0]*block_sizes[r,1]
        p.append(len(sample_wall[np.where(sample_wall==r)])/block_area)

    p = p/np.sum(p)
    return p

def corner_detect(arr):
    """
    Finds the corners of array arr.
    Input:
        arr: H x W arr
    Output:
        corners: H x W arr with corners highlighted
    """
    
    kernel = [
        [0,0,0],
        [1,1,1],
        [0,2,0]]
    kernel_side = [
        [0,1,0],
        [0,1,2],
        [0,1,0]]
    
    convo_mode = 'valid'
    padded_arr = np.zeros((arr.shape[0]+6, arr.shape[1]+6))
    padded_arr[3:-3, 3:-3] = arr
    corners = signal.convolve2d(padded_arr, kernel, mode=convo_mode)
    corners_side = signal.convolve2d(padded_arr, kernel_side, mode=convo_mode)
    
    corners = corners**2
    corners_side = corners_side**2
    
    corners = corners[2:-2, 2:-2]
    maxval = np.amax(corners)
    corners_side = corners_side[2:-2, 2:-2]
    maxval_side = np.amax(corners_side)
    
    corners[np.where(corners<maxval)]=0
    corners = corners>0
    corners = corners.astype(np.int32)
    corners_side[np.where(corners_side<maxval_side)]=0
    corners_side = corners_side>0
    corners_side = corners_side.astype(np.int32)
    
    corners = corners + corners_side
    corners[np.where(corners>1)] = 1
    
    return corners

def edge_to_block_coords(nodes_coords):
    """
    Shift node_coords to match the block's position in a figure.
    Input:
        nodes_coords: Nnodes x 2
    Output:
        nodes_coords_on_blocks: Nnodes x 2
    """
    nodes_coords_on_blocks = nodes_coords.copy()
    nodes_coords_on_blocks[:,0] = nodes_coords_on_blocks[:,0]+1/2
    nodes_coords_on_blocks[:,1] = nodes_coords_on_blocks[:,1]-1/2
    return nodes_coords_on_blocks

def block_to_edge_coords(nodes_coords):
    """
    Reverse edge_to_block_coords.
    Input:
        nodes_coords_on_blocks: Nnodes x 2
    Output:
        nodes_coords: Nnodes x 2
    """
    nodes_coords_on_edges = nodes_coords.copy()
    nodes_coords_on_edges[:,0] = nodes_coords_on_edges[:,0]-1/2
    nodes_coords_on_edges[:,1] = nodes_coords_on_edges[:,1]+1/2
    return nodes_coords_on_edges

def upscale(wall_sample, sr, sc):
    """
    Upscale wall_sample.
    Inputs:
        wall_sample: H x W array
        sr: scaling of rows
        sc: scaling of columns
    Outputs:
        upscaled_wall_sample: (sr*H) x (sc*W)
    """

    Nrows, Ncols  = wall_sample.shape
    
    upscaled_wall_sample = np.zeros((Nrows*sr, Ncols*sc))
    
    for i in range(Nrows):
        for j in range(Ncols):
            upscaled_wall_sample[(sr*i):(sr*(i+1)), (sc*j):(sc*(j+1))] = wall_sample[i, j]
            
    return upscaled_wall_sample

def upscale_space(space_sample, sr, sc, sd):
    """
    Upscale space_sample.
    Inputs:
        space_sample: H x W x D array
        sr: scaling of rows
        sc: scaling of columns
        sd: scaling of depth
    Outputs:
        upscaled_space_sample: (sr*H) x (sc*W) x (sd*D)
    """

    Nrows, Ncols, Ndepth = space_sample.shape
    upscaled_space_sample = np.zeros((Nrows*sr, Ncols*sc, Ndepth*sd))
    
    for i in range(Nrows):
        for j in range(Ncols):
            for k in range(Ndepth):
                upscaled_space_sample[(sr*i):(sr*(i+1)), (sc*j):(sc*(j+1)), (sd*k):(sd*(k+1))] = space_sample[i, j, k]
            
    return upscaled_space_sample

def get_corners_and_edges(target_wall):
    """
    Get corners and edges of the target_wall.
    Input:
        target_wall: H x W array
    Outputs:
        corners_and_edges: H x W array with corners and edges
        corners: H x W array with corners
    """

    # Compute gradients
    convo_mode = 'same'
    dx = np.abs(signal.convolve2d(target_wall, [[-1,1]], mode=convo_mode))>0
    dy = np.abs(signal.convolve2d(target_wall, [[-1], [1]], mode=convo_mode))>0

    dx = dx.astype(np.int32)
    dy = dy.astype(np.int32)

    wall = dx + dy + ((target_wall-np.amin(target_wall))/(np.amax(target_wall)-np.amin(target_wall)))

    edges = dx + dy
    expanded_edges = np.zeros((edges.shape[0]+1, edges.shape[1]))
    expanded_edges[:-1, :] = edges
    expanded_edges[-1, :] = 1
    edges = expanded_edges[:,1:]

    norm_target = (target_wall - np.mean(target_wall))/np.std(target_wall)

    corners_aux = np.zeros(edges.shape)
    corners_aux[np.where(edges==np.amax(edges))]=1
    flat_edges = edges.copy()
    flat_edges[np.where(flat_edges>1)]=1
    other_detected = np.abs(corner_detect(flat_edges))

    corners_aux = corners_aux + other_detected
    corners_aux[np.where(corners_aux>1)] = 1

    corners_and_edges = corners_aux+flat_edges
    corner_val = np.amax(corners_and_edges)
    corners = (corners_and_edges)>=corner_val
    corners = corners.astype(np.int32)
    
    return corners_and_edges, corners

def get_wall_graph(corners_and_edges, corners_mod):
    """
    Get graph composed of the block nodes and edges.
    Inputs:
        corners_and_edges: H x W array with corners and edges
        corners: H x W array with corners
    Outputs:
        G: graph of nodes(corners) and edges
        nodes_coords: coordinates of nodes
        coords_order: mapping of node id to row in nodes_coords
    """
    
    nodes0r, nodes0c = np.where(corners_mod>=1)
    nodes_num = -1*np.ones(corners_mod.shape)
    nodes_num_test = -1*np.ones(corners_and_edges.shape)
    levels = np.unique(nodes0r)
    corner_val = np.amax(corners_and_edges)
    sorted_levels = np.sort(levels)
    delta_sorted_levels = levels[1:] - levels[:-1]
    delta_level = np.amin(delta_sorted_levels)
    
    # Build the graph
    G = None
    G = nx.Graph()
    N = len(nodes0r) # number of nodes
    nodes_coords = np.zeros((1,2))
    n = 0
    nn = -1
    nodefrom = 0
    nodeto = 0
    coords_order = []

    for vi, level in enumerate(levels):
        nodesx_in_row = nodes0c[np.where(nodes0r==level)]
        for vj, nodex in enumerate(nodesx_in_row):
            if nodes_num_test[level, nodex]==-1:
                nn = nn + 1
                nodefrom = nn
                nodes_coords = np.vstack((nodes_coords, [nodex, level]))
                coords_order.append(nn)
                nodes_num_test[level, nodex] = nodefrom
            else:
                nodefrom = nodes_num_test[level, nodex]
            
            if (vi<len(levels)-1) and (vi>0): # If not first or last row, connect sideways
                if (corners_and_edges[level, nodex+1]>0) and (corners_and_edges[level, nodex+1]<corner_val):
                    to_the_right = corners_and_edges[level, (nodex+1):]
                    if len(to_the_right)>0 and (to_the_right==corner_val).any():
                        deltax_max = np.argmax(to_the_right==corner_val)
                        if nodes_num_test[level, nodex+deltax_max+1]==-1:
                            nn = nn + 1
                            nodes_coords = np.vstack((nodes_coords, [nodex+deltax_max+1, level]))
                            coords_order.append(nn)
                            nodeto = nn
                            nodes_num_test[level, nodex+deltax_max+1] = nodeto
                        else:
                            nodeto = nodes_num_test[level, nodex+deltax_max+1]

                        # Connect to the right node, if not first or last row
                        G.add_edge(nodefrom, nodeto, weight=deltax_max+1)

            # Vertical down
            to_below = corners_and_edges[(level+1):, nodex]
            if (vi<len(levels)-1):
                if (corners_and_edges[level+1, nodex]>0) and (corners_and_edges[level+1, nodex]<corner_val):
                    if len(to_below)>0 and (to_below==corner_val).any():
                        deltay_max = np.argmax(to_below==corner_val)
                        Ndeltas = int((deltay_max + 1)/delta_level)
                        for k in range(Ndeltas):
                            deltay = (k+1)*delta_level
                            if nodes_num_test[level + deltay, nodex]==-1:
                                nn = nn + 1
                                nodes_coords = np.vstack((nodes_coords, [nodex, level + deltay]))
                                coords_order.append(nn)
                                nodeto = nn
                                nodes_num_test[level + deltay, nodex] = nodeto
                            else:
                                nodeto = nodes_num_test[level + deltay, nodex]

                            # Connect to below node
                            G.add_edge(nodefrom, nodeto, weight=deltay)
                            nodefrom = nodeto
    nodes_coords = nodes_coords[1:, :]
    return G, nodes_coords, coords_order

def get_shortest_paths(G, nodes_coords, coords_order, target_wall, shortest_crack_lengths=None):
    """
    Find all shortest paths in G from top to bottom nodes.
    Inputs:
        G: graph of nodes(corners) and edges
        nodes_coords: coordinates of nodes
        coords_order: mapping of node id to row in nodes_coords
        target_wall: H x W array
        Ls: list of path lengths
        shortest_crack_lengths: list of shortest path lengths    
    Outputs:
        shortestG_list: list of graphs of the shortest path
        shortest_pos_list: list of dicts {node_idx: node_coords} of all shortest paths
        shortest_pos_list_ordered: list of arrays Nnodes x 2
        nodelist_list: list of mappings from array row into node idxs
        pos: dict {node_idx: node_coords} of all nodes
        shortest_crack_lengths: list of shortest path lengths
    """

    A = nx.to_numpy_array(G)

    if shortest_crack_lengths is None:
        shortest_crack_lengths = []
    
    if len(A)==0:
        print("Empty graph.")
    
    Lcount = []
    for i in range(A.shape[0]):
        if (A[i,i:][::-1]>0).any():
            Lcount.append(A.shape[1]-i - np.argmax(A[i,i:][::-1]>0))
    
    L = np.amax(Lcount)
    D = np.zeros((A.shape[0], L-1))
    
    for i in range(A.shape[0]-1):
        Ad = A[i, (i+1):(i+L)]
        D[i, :len(Ad)] = Ad

    nodes_coords_on_blocks = edge_to_block_coords(nodes_coords)
    pos = {}
        
    for k in coords_order:
        coord = nodes_coords_on_blocks[k, :].copy()
        pos.update({k: coord})
        
    bottom_level = np.amin(nodes_coords_on_blocks[:,1])
    top_level = np.amax(nodes_coords_on_blocks[:,1])
    bottom_nodes = np.where(nodes_coords_on_blocks[:, 1]==bottom_level)[0]
    top_nodes = np.where(nodes_coords_on_blocks[:, 1]==top_level)[0]
    
    shortest_paths = []
    for bottom_node in bottom_nodes:
        for top_node in top_nodes:
            if G.has_node(bottom_node) and G.has_node(top_node):
                try:
                    path_length = nx.shortest_path_length(G, bottom_node, top_node, weight="weight")
                    shortest_paths.append([bottom_node, top_node, path_length])
                except:
                    print("Error in computing path length -- ")
                    print('Bottom:', bottom_node, 'Top:', top_node)
                    continue

    shortest_paths = np.array(shortest_paths)
    min_shortest_paths = np.amin(shortest_paths[:,2])
    shortest_crack_lengths.append(min_shortest_paths/target_wall.shape[0])
    min_idxs = np.where(shortest_paths[:,2]==min_shortest_paths)[0]
    shortestG_list = []
    shortest_pos_list = []
    shortest_pos_list_ordered = []
    nodelist_list = []

    for min_idx in min_idxs:
        try:
            nodelist = np.array(nx.shortest_path(G, shortest_paths[min_idx, 0], shortest_paths[min_idx, 1], weight="weight")).astype(np.int32)
            Nshort = len(nodelist)
            shortestG = nx.Graph()

            for i in range(0, Nshort-1):
                edge_weight = G.get_edge_data(nodelist[i], nodelist[i+1])["weight"]
                shortestG.add_edge(nodelist[i], nodelist[i+1], weight=edge_weight)

            shortest_coords = nodes_coords_on_blocks[nodelist, :].copy()
            shortest_pos = {}
            for k, coord in enumerate(shortest_coords):
                shortest_pos.update({nodelist[k]: coord})

            shortestG_list.append(shortestG)
            shortest_pos_list.append(shortest_pos)
            shortest_pos_list_ordered.append(shortest_coords)
            nodelist_list.append(nodelist)
        except:
            print("Error in building the min graph section.")
            print("All idxs:", min_idxs)
            print("Cur idx:", min_idx)
            continue

    return shortestG_list, shortest_pos_list, shortest_pos_list_ordered, nodelist_list, pos, shortest_crack_lengths


def plot_wall(G, pos, shortestG_list, shortest_pos_list, target_wall, colors=None, with_labels=False, round_values=True, node_size=1, graph_color='lightblue', ax=None):
    """
    Plot a wall with the corresponding graph and highlights the 'crack path' or shortest path.
    Inputs:
        G: graph of nodes(corners) and edges
        shortestG_list: list of graphs of the shortest path
        shortest_pos_list: list of dicts {node_idx: node_coords} of all shortest paths
        target_wall: H x W array
        colors: Ncolors x 3 array for the shortest paths
        with_labels: whether to plot or not the node labels
        round_values: whether to round the target_wall values
        node_size: graph node sizes
        graph_color: color of the graph G
        ax: axis in which to plot
    """

    if ax is None:
        fig, (ax) = plt.subplots(1,1, dpi=dpi)

    font_size = 3
    width = 0.25
    vmin = -1.25
    vmax = 15
    cmap = cm.pink
    ax.set_facecolor("lightblue")
    if round_values:
        ax.imshow(np.round(target_wall), cmap=cmap, vmin=vmin, vmax=vmax)
    else:
        ax.imshow(target_wall, cmap=cmap, vmin=vmin, vmax=vmax)

    nx.draw_networkx(G, pos=pos,  with_labels=with_labels, ax=ax, node_size=node_size, node_color=graph_color, edge_color=graph_color,font_size=font_size, font_color='b', alpha=0.95, linewidths=0, width=width)
    if colors is None:
        colors = cm.hsv(np.linspace(0., 0.9, len(shortestG_list)))
    for j, shortest_data in enumerate(zip(shortestG_list, shortest_pos_list)):
        shortestG, shortest_pos = shortest_data
        short_path_color = colors[j, :][np.newaxis, :]
        nx.draw_networkx(shortestG, with_labels=with_labels, pos=shortest_pos, ax=ax, node_size=node_size, node_color=short_path_color, edge_color=short_path_color, font_size=font_size, alpha=1, font_color='k', linewidths=0, width=width)

    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)

    plt.axis('off')
    plt.tight_layout()
    plt.gcf().set_size_inches(imgw//2, imgh//2)
    plt.draw()
    plt.pause(0.001)
    
def noisify_sample(sample_blocks, block_sizes):
    """
    Introduce random noise into a sample block to distinguish adjacent blocks of the same type
    Inputs:
        sample_blocks: H x W array
        block_sizes: Nblocks x 3 array
    Outputs:
        noisy_sample_blocks: H x W array
    """

    Nr, Nc = sample_blocks.shape
    noisy_sample_blocks = np.zeros(sample_blocks.shape)
    
    for i in range(Nr):
        for j in range(Nc):
            block_type = sample_blocks[i, j]
            block_size = block_sizes[block_type, :]
            if noisy_sample_blocks[i, j]==0: 
                noisy_sample_blocks[i:(i+block_size[1]), j:(j+block_size[0])] = block_type + (np.random.rand()-0.5)
                
    return noisy_sample_blocks

def create_target_space(target_size, fill_opts):
    """
    Creates a padded, empty space of target size.
    Inputs:
        target_size: (H, W, D) tuple
        fill_opts: filling options
    Output:
        target_space: padded H x W x D
    """
    Ntarget_rows, Ntarget_cols, Ntarget_depth = target_size
    s = fill_opts["scale"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]
    target_space = -1*np.ones((Ntarget_rows+row_pad, s*header_pad + Ntarget_cols+col_pad, Ntarget_depth + depth_pad))
    return target_space
    
def remove_target_pad(target_space, fill_opts):
    """
    Remove padding from a target space.
    Inputs:
        target_space: H x W x D array
        fill_opts: filling options
    Outputs:
        target_space: unpadded target_space
    """
    s = fill_opts["scale"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]
    target_space = target_space[:-row_pad, (s*header_pad):-col_pad, :-depth_pad]
    return target_space

def prefil_padded_target_space(target_wall, target_space, k, fill_opts):
    """
    Prefil a padded target space kth wall with target_wall.
    Inputs:
        target_wall: H x W array
        target_space: padded target_space
        k: leaf index
        fill_opts: filling options
    Outputs:
        new_target_space: prefilled padded target_space
    """
    s = fill_opts["scale"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]

    new_target_space = target_space.copy()

    col_step = s*header_pad
    new_target_space[:-row_pad, col_step:-col_pad, k] = target_wall

    return new_target_space

def get_new_block_id(blockN, block_ids):
    """
    Randomly generates a block id not in block_ids.
    Inputs:
        blockN: the block type (integer)
        block_ids: list of existing block_ids
    Output:
        new_block_id: the new block id
    """
    unique_id = False
    while not unique_id:
        new_block_id = blockN + 0.25*np.random.rand()

        if (block_ids!=new_block_id).all():
            unique_id = True
    return new_block_id

def get_surrounding_blocks(target_space, s, i, j, k):
    """
    Get blocks surrounding a header position.
    Inputs:
        target_space: H x W x D
        s: scaling factor
        i: row
        j: col
        k: depth
    Outputs:
        surrounding_blocks: array of surrounding blocks ids
    """

    Nrows, Ncols, Ndepth = target_space.shape
    surrounding_blocks = np.array([])
    
    # above row
    top_row = target_space[i-1, (j-1):(j+s+1), k]
    surrounding_blocks = np.append(surrounding_blocks, top_row)

    if (i < Nrows):
        bottom_row = target_space[i+1, (j-1):(j+s+1), k]
        surrounding_blocks = np.append(surrounding_blocks, bottom_row)
    
    left_block = target_space[i, j-1, k]
    surrounding_blocks = np.append(surrounding_blocks, left_block)
    
    if (j + s + 1) < Ncols:
        right_block = target_space[i, j+s+1, k]
        surrounding_blocks = np.append(surrounding_blocks, right_block)
    
    surrounding_blocks = np.unique(np.round(surrounding_blocks)).astype(np.int32)
    return surrounding_blocks

def fill_verticals_aux(target_space, aux_target_space, all_verticals, side_verticals, Nblock_types, max_block_ratio, walls_block_probs, block_sizes, block_ids, no_adj, fill_opts, fill_unit_opt=True, pick_random=False):
    """
    Auxiliary function for fill_verticals.
    """

    Ntarget_rows, Ntarget_cols, Ntarget_depth  = target_space.shape

    d = [-1, 1]
    s = fill_opts["scale"]
    header_height = fill_opts["header_height"]
    header_depth = fill_opts["header_depth"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]
    lintel_id = fill_opts["lintel_id"]

    for i in range(0, Ntarget_rows-row_pad, s*header_height):
        for k in range(0, Ntarget_depth-depth_pad, header_depth):
            block_probs = walls_block_probs[k]
            steps = np.where(side_verticals[i, :, k]==1.)[0] + 1
            steps = [steps[0]]
            for step in steps:
                ahead_verticals = np.where(all_verticals[i, step:, k]==1.)[0]
                ahead_vertical = ahead_verticals[0]
                ahead_value = aux_target_space[i, step+ahead_vertical, k]
                allowed_blocks = np.arange(0, Nblock_types)
                if ahead_value==0: # If empty space ahead, then limit the size of the blocks to be used
                    max_block_size = np.round(max_block_ratio*ahead_vertical)
                    allowed_blocks = np.where(block_sizes[:,0]<=max_block_size)[0]

                if target_space[i, step, k]==-1: # If the header position is free
                    #-- block selection logic
                    if i ==0: # If it is the first row, pick at rates proportional to sample wall
                        possible_fill_blocks = None
                        possible_choices = np.arange(0, Nblock_types)
                        possible_choice_probs = block_probs.copy()
                        prev_block = np.round(target_space[i,  step-1, k])
                        if (no_adj==prev_block).any():
                            valid = np.where(possible_choices!=prev_block)
                            possible_choices = possible_choices[valid]
                            possible_choice_probs = possible_choice_probs[valid]
                            possible_choice_probs = possible_choice_probs/np.sum(possible_choice_probs)

                        blockN = np.random.choice(possible_choices, p=possible_choice_probs)
                        block_size = block_sizes[blockN, :]
                    else:
                        # Avoid placing adjacent blocks of specified types 
                        invalid_blocks = []
                        surrounding_blocks = get_surrounding_blocks(target_space, s, i, step, k)

                        for block_type in no_adj:
                            if (surrounding_blocks==block_type).any():
                                invalid_blocks.append(block_type)
                        invalid_blocks = np.array(invalid_blocks)
                        # If it is not the first row, pick a segment of the row above
                        # near_blocks = target_space[i-1,  step:(step+longest_width+1), k]
                        top_near_blocks = target_space[i-s, step:, k]

                        # Find where edges lie
                        top_edges = np.convolve(top_near_blocks, d, mode='valid')
                        top_edges = top_edges[(s-1)::s]

                        # All sizes that do not match with an edge are considered valid
                        top_possible_sizes = s*(np.where(top_edges==0)[0] + 1)

                        bottom_near_blocks = target_space[i + s*header_height + 1, step:, k]
                        bottom_edges = np.convolve(bottom_near_blocks, d, mode='valid')
                        bottom_edges = bottom_edges[(s-1)::s]
                        bottom_possible_sizes = s*(np.where(bottom_edges==0)[0] + 1)

                        possible_sizes = np.intersect1d(top_possible_sizes, bottom_possible_sizes)

                        # Keep only available sizes
                        available_sizes = block_sizes[:, 0]
                        possible_fill_blocks = np.array([])
                        for vi, v in enumerate(possible_sizes):
                            available_blocks = np.where(available_sizes==v)[0]
                            if len(available_blocks) > 0:
                                possible_fill_blocks = np.append(possible_fill_blocks, available_blocks)

                        possible_fill_blocks = np.intersect1d(possible_fill_blocks, allowed_blocks)
                        possible_fill_blocks = possible_fill_blocks.astype(np.int32)
                        if len(invalid_blocks)>0:
                            for invalid_block in invalid_blocks:
                                possible_fill_blocks = possible_fill_blocks[np.where(possible_fill_blocks!=invalid_block)]

                        if len(possible_fill_blocks)>0: # If there are any possible blocks
                            if np.sum(block_probs[possible_fill_blocks]) > 0: # If those blocks are not being supressed
                                possible_fill_blocks_probs = block_probs[possible_fill_blocks]/np.sum(block_probs[possible_fill_blocks])
                                blockN = np.random.choice(possible_fill_blocks, p=possible_fill_blocks_probs) # Pick one
                                block_size = block_sizes[blockN, :]
                        elif not pick_random:
                            print("Failed to get a block on the first try")
                        else:
                            # If there are no valid sizes, just pick randomly, but supress unit voxels
                            possible_choices = np.arange(1, Nblock_types-1)
                            possible_block_probs = block_probs[possible_choices]
                            possible_block_probs = possible_block_probs/np.sum(possible_block_probs)
                            blockN = np.random.choice(possible_choices, p=possible_block_probs)
                            block_size = block_sizes[blockN, :]
                    #--

                    if (step + block_size[0] <= Ntarget_cols) and (i + block_size[1] <= Ntarget_rows+s):
                        # check if space is empty
                        if (target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])]==-1).all():
                            new_block_id = get_new_block_id(blockN, block_ids)
                            blocks_ids = np.append(block_ids, new_block_id)
                            target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id
                        else:
                            # if space is not empty, we check what is the available space, for now we only check orthogonally
                            if (target_space[i, step:, k]==-1).all():
                                x_gap = len(target_space[i, step:, k])
                            else:
                                x_gap = np.argmin(target_space[i, step:, k]==-1)

                            if (target_space[i:, step, k]==-1).all():
                                y_gap = len(target_space[i:, step, k])
                            else:
                                y_gap = np.argmin(target_space[i:, step, k]==-1)

                            if (target_space[i, step, k:]==-1).all():
                                z_gap = len(target_space[i, step, k:])
                            else:
                                z_gap = np.argmin(target_space[i, step, k:]==-1)

                            if possible_fill_blocks is None:
                                possible_fill_blocks = np.where((block_sizes[:,0]<=x_gap)&(block_sizes[:,1]<=y_gap)&(block_sizes[:,2]<=z_gap))[0]
                            else:
                                possible_gap_fill_blocks = np.where((block_sizes[:,0]<=x_gap)&(block_sizes[:,1]<=y_gap)&(block_sizes[:,2]<=z_gap))[0]
                                possible_fill_blocks = np.intersect1d(possible_fill_blocks, possible_gap_fill_blocks)
                                if len(possible_fill_blocks)==0:
                                    possible_fill_blocks = possible_gap_fill_blocks

                            if len(possible_fill_blocks)>0:
                                probs_fill_blocks = block_probs[possible_fill_blocks]
                                if len(probs_fill_blocks) > 1 and np.sum(probs_fill_blocks) > 0:
                                    probs_fill_blocks = probs_fill_blocks/np.sum(probs_fill_blocks)
                                    blockN = np.random.choice(possible_fill_blocks, p=probs_fill_blocks)
                                else:
                                    blockN = possible_fill_blocks[0]

                                block_size = block_sizes[blockN, :]
                                if (target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])]==-1).all():
                                    new_block_id = get_new_block_id(blockN, block_ids)
                                    blocks_ids = np.append(block_ids, new_block_id)
                                    target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id
                                elif (target_space[i:(i+s), step:(step + s), k:(k+1)]==-1).all() and fill_unit_opt:
                                    # Try to fill current position with specified unit voxel
                                    if unit_blockN==-1:
                                        blockN = block_sizes.shape[0]-1
                                    elif unit_blockN==-2:
                                        if verbose:
                                            print(i, "/", Ntarget_rows, step, k, "| Place is already occupied >> Inserting unit block available.")
                                        continue
                                    else:
                                        blockN=unit_blockN

                                    block_size = block_sizes[blockN, :]
                                    new_block_id = get_new_block_id(blockN, block_ids)
                                    blocks_ids = np.append(block_ids, new_block_id)
                                    target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id

                                    if verbose:
                                        print(i, "/", Ntarget_rows, step, k, "| Place is already occupied >> Inserting unit block available.")
                            else:
                                if verbose:
                                    print(i, "/", Ntarget_rows, step, k, "| Place is already occupied >> No alternative block available.")
    return target_space, aux_target_space, all_verticals, block_ids

def fill_verticals(target_space, sample_blocks, block_sizes, walls_block_probs, fill_opts, block_ids=np.array([]), verbose=True, no_adj=np.array([0,2]), unit_blockN=-2):
    """
    Fill target_space vertical edges of openings (e.g., sides of doors) and laterals of the target space.
    Inputs:
        target_space: H x W x D array
        sample_blocks: Hs x Ws array
        block_sizes: Nblocks x 3 array
        walls_block_probs: Nblocks array
        fill_opts: filling options
        block_ids: array of block instance ids
        no_adj: array with block ids that should not be adjacent
        unit_blockN: the unit voxel block type
    Outputs:
        target_space: H x W x D array
        block_ids: array of block instance ids
    """
    Ntarget_rows, Ntarget_cols, Ntarget_depth  = target_space.shape

    Nblock_types = block_sizes.shape[0]
    longest_width = np.amax(block_sizes[:,0])

    d = [-1, 1]
    s = fill_opts["scale"]
    header_height = fill_opts["header_height"]
    header_depth = fill_opts["header_depth"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]
    lintel_id = fill_opts["lintel_id"]

    aux_target_space = target_space.copy()
    aux_target_space[np.where(aux_target_space<-1)] = 0

    right_verticals = aux_target_space.copy()
    right_verticals[:, :-1, :] = right_verticals[:, :-1, :] - right_verticals[:, 1:, :]
    right_verticals[np.where(right_verticals<0)] = 0
    right_verticals[np.where(right_verticals>1)] = 1

    left_verticals = aux_target_space.copy()
    left_verticals = left_verticals[:, ::-1, :]
    left_verticals[:, :-1, :] = left_verticals[:, :-1, :] - left_verticals[:, 1:, :]
    left_verticals[np.where(left_verticals<0)] = 0
    left_verticals[np.where(left_verticals>1)] = 1

    all_verticals = right_verticals + left_verticals[:, ::-1, :]

    # Fill left major verticals
    target_space = target_space[:,::-1,:]
    aux_target_space = aux_target_space[:,::-1,:]
    all_verticals = all_verticals[:, ::-1, :]
    plot_count = 0
    plot_rate = 10
    plot_pause = 0.01
    left_max_block_ratio = 0.7 # Specifies the largest block size that can be used as a ratio of the available gap
    right_max_block_ratio = 0.7
    
    target_space, aux_target_space, all_verticals, block_ids = fill_verticals_aux(target_space, aux_target_space, all_verticals, left_verticals, Nblock_types, left_max_block_ratio, walls_block_probs, block_sizes, block_ids, no_adj, fill_opts)

    # Fill right major verticals
    target_space = target_space[:,::-1,:]
    aux_target_space = aux_target_space[:,::-1,:]
    all_verticals = all_verticals[:, ::-1, :]
    target_space, aux_target_space, all_verticals, block_ids = fill_verticals_aux(target_space, aux_target_space, all_verticals, right_verticals, Nblock_types, right_max_block_ratio, walls_block_probs, block_sizes, block_ids, no_adj, fill_opts, fill_unit_opt=False)

    # Fill left minor verticals
    target_space = target_space[:,::-1,:]
    aux_target_space = aux_target_space[:,::-1,:]
    all_verticals = all_verticals[:, ::-1, :]
    target_space, aux_target_space, all_verticals, block_ids = fill_verticals_aux(target_space, aux_target_space, all_verticals, left_verticals, Nblock_types, left_max_block_ratio, walls_block_probs, block_sizes, block_ids, no_adj, fill_opts, pick_random=True)
    

    # Fill right minor verticals
    target_space = target_space[:,::-1,:]
    aux_target_space = aux_target_space[:,::-1,:]
    all_verticals = all_verticals[:, ::-1, :]
    target_space, aux_target_space, all_verticals, block_ids = fill_verticals_aux(target_space, aux_target_space, all_verticals, right_verticals, Nblock_types, right_max_block_ratio, walls_block_probs, block_sizes, block_ids, no_adj, fill_opts, fill_unit_opt=False, pick_random=True)

    return target_space, block_ids

def fill_space(target_space, sample_blocks, block_sizes, walls_block_probs, fill_opts, block_ids=np.array([]), verbose=True, no_adj=np.array([0,2]), unit_blockN=-2):
    """
    Fill target_space vertical edges of openings (e.g., sides of doors) and laterals of the target space.
    Inputs:
        target_space: H x W x D array
        sample_blocks: Hs x Ws array
        block_sizes: Nblocks x 3 array
        walls_block_probs: Nblocks array
        fill_opts: filling options
        block_ids: array of block instance ids
        no_adj: array with block ids that should not be adjacent
        unit_blockN: the unit voxel block type
    Outputs:
        target_space: H x W x D array
    """

    Ntarget_rows, Ntarget_cols, Ntarget_depth = target_space.shape

    fill_blocks_list = np.array([10, 11])

    Nblock_types = block_sizes.shape[0]
    longest_width = np.amax(block_sizes[:,0])

    d = [-1, 1]
    s = fill_opts["scale"]
    header_height = fill_opts["header_height"]
    header_depth = fill_opts["header_depth"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]
    lintel_id = fill_opts["lintel_id"]

    for i in range(0, Ntarget_rows-row_pad, s*header_height):
        for k in range(0, Ntarget_depth-depth_pad, header_depth):
            block_probs = walls_block_probs[k]
            step = s*header_pad
            while step < Ntarget_cols:
                if target_space[i, step, k]==-1: # If the header position is free
                    #-- block selection logic
                    if i ==0: # If it is the first row, pick at rates proportional to sample wall
                        possible_fill_blocks = None
                        possible_choices = np.arange(0, Nblock_types)
                        possible_choice_probs = block_probs.copy()
                        prev_block = np.round(target_space[i,  step-1, k])
                        if (no_adj==prev_block).any():
                            valid = np.where(possible_choices!=prev_block)
                            possible_choices = possible_choices[valid]
                            possible_choice_probs = possible_choice_probs[valid]
                            possible_choice_probs = possible_choice_probs/np.sum(possible_choice_probs)

                        blockN = np.random.choice(possible_choices, p=possible_choice_probs)
                        block_size = block_sizes[blockN, :]
                    else:
                        # Avoid placing adjacent blocks of specified types 
                        invalid_blocks = []
                        surrounding_blocks = get_surrounding_blocks(target_space, s, i, step, k)

                        for block_type in no_adj:
                            if (surrounding_blocks==block_type).any():
                                invalid_blocks.append(block_type)
                        invalid_blocks = np.array(invalid_blocks)
                        # If it is not the first row, pick a segment of the row above
                        near_blocks = target_space[i-1,  step:, k]

                        # Find where edges lie
                        edges = np.convolve(near_blocks, d, mode='valid')
                        edges = edges[(s-1)::s]

                        # All sizes that do not match with an edge are considered valid
                        possible_sizes = s*(np.where(edges==0)[0] + 1)

                        # Keep only available sizes
                        available_sizes = block_sizes[:, 0]
                        possible_fill_blocks = np.array([])
                        for vi, v in enumerate(possible_sizes):
                            available_blocks = np.where(available_sizes==v)[0]
                            if len(available_blocks) > 0:
                                possible_fill_blocks = np.append(possible_fill_blocks, available_blocks)

                        possible_fill_blocks = possible_fill_blocks.astype(np.int32)
                        if len(invalid_blocks)>0:
                            for invalid_block in invalid_blocks:
                                possible_fill_blocks = possible_fill_blocks[np.where(possible_fill_blocks!=invalid_block)]

                        if len(possible_fill_blocks)>0: # If there are any possible blocks
                            if np.sum(block_probs[possible_fill_blocks]) > 0: # If those blocks are not being supressed
                                possible_fill_blocks_probs = block_probs[possible_fill_blocks]/np.sum(block_probs[possible_fill_blocks])
                                blockN = np.random.choice(possible_fill_blocks, p=possible_fill_blocks_probs) # Pick one
                                block_size = block_sizes[blockN, :]
                        else:
                            # If there are no valid sizes, just pick randomly, but supress unit voxels
                            possible_choices = np.arange(1, Nblock_types-1)
                            possible_block_probs = block_probs[possible_choices]
                            possible_block_probs = possible_block_probs/np.sum(possible_block_probs)
                            blockN = np.random.choice(possible_choices, p=possible_block_probs)
                            block_size = block_sizes[blockN, :]

                    if (step + block_size[0] <= Ntarget_cols) and (i + block_size[1] <= Ntarget_rows+s):
                        # check if space is empty
                        if (target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])]==-1).all():
                            new_block_id = get_new_block_id(blockN, block_ids)
                            blocks_ids = np.append(block_ids, new_block_id)
                            target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id
                        else:
                            if (target_space[i, step:, k]==-1).all():
                                x_gap = len(target_space[i, step:, k])
                            else:
                                x_gap = np.argmin(target_space[i, step:, k]==-1)

                            if (target_space[i:, step, k]==-1).all():
                                y_gap = len(target_space[i:, step, k])
                            else:
                                y_gap = np.argmin(target_space[i:, step, k]==-1)

                            if (target_space[i, step, k:]==-1).all():
                                z_gap = len(target_space[i, step, k:])
                            else:
                                z_gap = np.argmin(target_space[i, step, k:]==-1)

                            if possible_fill_blocks is None:
                                possible_fill_blocks = np.where((block_sizes[:,0]<=x_gap)&(block_sizes[:,1]<=y_gap)&(block_sizes[:,2]<=z_gap))[0]
                            else:
                                possible_gap_fill_blocks = np.where((block_sizes[:,0]<=x_gap)&(block_sizes[:,1]<=y_gap)&(block_sizes[:,2]<=z_gap))[0]
                                possible_fill_blocks = np.intersect1d(possible_fill_blocks, possible_gap_fill_blocks)
                                if len(possible_fill_blocks)==0:
                                    possible_fill_blocks = possible_gap_fill_blocks

                            if len(possible_fill_blocks)>0:
                                probs_fill_blocks = block_probs[possible_fill_blocks]
                                if len(probs_fill_blocks) > 1 and np.sum(probs_fill_blocks) > 0:
                                    probs_fill_blocks = probs_fill_blocks/np.sum(probs_fill_blocks)
                                    blockN = np.random.choice(possible_fill_blocks, p=probs_fill_blocks)
                                else:
                                    blockN = possible_fill_blocks[0]

                                block_size = block_sizes[blockN, :]
                                if (target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])]==-1).all():
                                    new_block_id = get_new_block_id(blockN, block_ids)
                                    blocks_ids = np.append(block_ids, new_block_id)
                                    target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id
                                elif (target_space[i:(i+s), step:(step + s), k:(k+1)]==-1).all():
                                    # Try to fill current position with specified unit voxel
                                    if unit_blockN==-1:
                                        blockN = block_sizes.shape[0]-1
                                    elif unit_blockN==-2:
                                        if verbose:
                                            print(i, "/", Ntarget_rows, step, k, "| Place is already occupied >> Moving header.")
                                        continue
                                    else:
                                        blockN=unit_blockN

                                    block_size = block_sizes[blockN, :]
                                    new_block_id = get_new_block_id(blockN, block_ids)
                                    blocks_ids = np.append(block_ids, new_block_id)
                                    target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id

                                    if verbose:
                                        print(i, "/", Ntarget_rows, step, k, "| Place is already occupied >> Inserting unit block available.")
                            else:
                                if verbose:
                                    print(i, "/", Ntarget_rows, step, k, "| Place is already occupied >> No alternative block available.")
                    else:
                        row_gap = Ntarget_cols
                        if row_gap > 1: # Only fill gaps larger than size 1
                            possible_fill_blocks = np.where((block_sizes[:,0]<=row_gap)&(block_sizes[:,1]==s*1))[0]
                            if len(possible_fill_blocks)>0:
                                probs_fill_blocks = block_probs[possible_fill_blocks]
                                if len(probs_fill_blocks) > 1 and np.sum(probs_fill_blocks) > 0:
                                    probs_fill_blocks = probs_fill_blocks/np.sum(probs_fill_blocks)
                                    blockN = np.random.choice(possible_fill_blocks, p=probs_fill_blocks)
                                else:
                                    blockN = possible_fill_blocks[0]

                                block_size = block_sizes[blockN, :]
                                new_block_id = get_new_block_id(blockN, block_ids)
                                blocks_ids = np.append(block_ids, new_block_id)
                                target_space[i:(i+block_size[1]), step:(step + block_size[0]), k:(k+block_size[2])] = new_block_id
                                
                            else:
                                if verbose:
                                    print(i, "/", Ntarget_rows, step, k, "| Insufficient space >> No alternative block available")
                        else:
                            if verbose:
                                print(i, "/", Ntarget_rows, step, k, "| Insufficient space >> Row gap too small")

                    # move the header
                    step = step + block_size[0]
                else: # If the header position is not free, move it to the next free position
                    where_empty = np.where(target_space[i, step:, k]==-1)[0]
                    if len(where_empty) > 0:
                        dstep = where_empty[0]
                        step = step + dstep
                    else:
                        step = step + 1

        target_space, block_ids = merge_blocks(target_space, block_ids, fill_blocks_list, block_sizes, fill_opts)
        target_space = target_space[:, ::-1, :]
        target_space, block_ids = merge_blocks(target_space, block_ids, fill_blocks_list, block_sizes, fill_opts)
        target_space = target_space[:, ::-1, :]

    target_space, block_ids = merge_blocks(target_space, block_ids, fill_blocks_list, block_sizes, fill_opts)
    target_space = target_space[:, ::-1, :]
    target_space, block_ids = merge_blocks(target_space, block_ids, fill_blocks_list, block_sizes, fill_opts)
    target_space = target_space[:, ::-1, :]

    return target_space

def get_complete_wall_graph(corners_and_edges, corners_mod):
    """
    Get graph of wall corners and edges including lateral, top, and bottoms.
    Inputs:
        corners_and_edges: H x W array with corners and edges
        corners: H x W array with corners
    Outputs:
        G: graph of nodes(corners) and edges
        nodes_coords: coordinates of nodes
        coords_order: mapping of node id to row in nodes_coords
    """

    corner_val = np.amax(corners_and_edges)
    corners_and_edges[:, 0] = corners_and_edges[:, 0] + 1
    corners_and_edges[:, -1] = corners_and_edges[:, -1] + 1
    corners_mod = (corners_and_edges)>=corner_val
    corners_mod = corners_mod.astype(np.int32)
    
    nodes0r, nodes0c = np.where(corners_mod>=1)
    nodes_num = -1*np.ones(corners_mod.shape)
    nodes_num_test = -1*np.ones(corners_and_edges.shape)
    levels = np.unique(nodes0r)
    
    W = corners_and_edges.shape[1]
    H = corners_and_edges.shape[0]
    
    sorted_levels = np.sort(levels)
    delta_sorted_levels = levels[1:] - levels[:-1]
    delta_level = np.amin(delta_sorted_levels)
    
    # Build the graph
    G = None
    G = nx.Graph()
    N = len(nodes0r) # number of nodes
    nodes_coords = np.zeros((1,2))
    n = 0
    nn = -1
    nodefrom = 0
    nodeto = 0
    coords_order = []
    for vi, level in enumerate(levels):
        
        nodesx_in_row = nodes0c[np.where(nodes0r==level)] # Get all nodes in the level
        for vj, nodex in enumerate(nodesx_in_row):
            if nodes_num_test[level, nodex]==-1:
                nn = nn + 1
                nodefrom = nn
                nodes_coords = np.vstack((nodes_coords, [nodex, level]))
                coords_order.append(nn)
                nodes_num_test[level, nodex] = nodefrom
            else:
                nodefrom = nodes_num_test[level, nodex]
            
            if nodex < W-1: # If not first or last row, connect sideways
                if (corners_and_edges[level, nodex+1]>0) and (corners_and_edges[level, nodex+1]<corner_val):
                    to_the_right = corners_and_edges[level, (nodex+1):]
                    if len(to_the_right)>0 and (to_the_right==corner_val).any():
                        deltax_max = np.argmax(to_the_right==corner_val)
                        deltax = deltax_max + 1
                            
                        if nodes_num_test[level, nodex+deltax]==-1:
                            nn = nn + 1
                            nodes_coords = np.vstack((nodes_coords, [nodex+deltax, level]))
                            coords_order.append(nn)
                            nodeto = nn
                            nodes_num_test[level, nodex+deltax] = nodeto
                        else:
                            nodeto = nodes_num_test[level, nodex+deltax]

                        # Connect to the right node, if not first or last row
                        G.add_edge(nodefrom, nodeto, weight=deltax)

            # Vertical down
            to_below = corners_and_edges[(level+1):, nodex]
            if (vi<len(levels)-1):
                if (corners_and_edges[level+1, nodex]>0) and (corners_and_edges[level+1, nodex]<corner_val):
                    if len(to_below)>0 and (to_below==corner_val).any():
                        deltay_max = np.argmax(to_below==corner_val)
                        Ndeltas = int((deltay_max + 1)/delta_level)
                        
                        for k in range(Ndeltas):
                            deltay = (k+1)*delta_level
                            if nodes_num_test[level + deltay, nodex]==-1:
                                nn = nn + 1
                                nodes_coords = np.vstack((nodes_coords, [nodex, level + deltay]))
                                coords_order.append(nn)
                                nodeto = nn
                                nodes_num_test[level + deltay, nodex] = nodeto
                            else:
                                nodeto = nodes_num_test[level + deltay, nodex]

                            # Connect to below node
                            G.add_edge(nodefrom, nodeto, weight=deltay)
                            nodefrom = nodeto

    nodes_coords = nodes_coords[1:, :]
    first_column = np.where(nodes_coords[:,0]==0)[0]
    last_column = np.where(nodes_coords[:,0]==W-1)[0]
    nodes_coords[first_column, 0] = nodes_coords[first_column, 0] - 1
    nodes_coords[last_column, 0] = nodes_coords[last_column, 0] + 1

    return G, nodes_coords, coords_order

def get_cross_neighbors(target_wall, x, y):
    """
    Get diagonal block ids around a node.
    Inputs:
        target_wall: H x W array
        x, y: node coordinates from node_coords
    Output:
        xblocks: list of cross neighboring block ids
    """
    # Gets the cross neighbors of a given point
    xblocks = []
    
    H = target_wall.shape[0]
    W = target_wall.shape[1]
    
    x = int(x)
    y = int(y)
    
    if x > 0 and y > 0:
        block = target_wall[y-1, x-1]
        xblocks.append(block)
    else:
        xblocks.append(-1)
        
    if x < W-1 and y < H-1:
        block = target_wall[y+1, x+1]
        xblocks.append(block)
    else:
        xblocks.append(-1)
        
    if x < W-1 and y > 0:
        block = target_wall[y-1, x+1]
        xblocks.append(block)
    else:
        xblocks.append(-1)
        
    if x > 0 and y < H-1:
        block = target_wall[y+1, x-1]
        xblocks.append(block)
    else:
        xblocks.append(-1)
        
    return xblocks

def get_neighboring_blocks_ids(target_wall, nodes_coords):
    """
    Get neighboring blocks for nodes.
    Inputs:
        target_wall: H x W array
        nodes_coords: Nnodes x 2 array
    Output:
        neighboring_block_ids: Nnodes x 4 array
    """
    
    neighboring_blocks_ids = []
    for j, coord in enumerate(nodes_coords):
        xblocks = get_cross_neighbors(target_wall, coord[0], coord[1])
        neighboring_blocks_ids.append(xblocks)
    neighboring_blocks_ids = np.array(neighboring_blocks_ids)

    return neighboring_blocks_ids

def get_wall_blocks(block_ids, neighboring_blocks_ids, Gcomplete):
    """
    Get wall blocks.
    Inputs:
        block_ids: list of block ids
        neighboring_blocks_ids: Nnodes x 4 array
        Gcomplete: graph of corners and edges of the wall
    Output:
        wall_blocks_list: list of dicts {block_id: block_graph}
    """

    block_nodes = {}
    Nnodes = neighboring_blocks_ids.shape[0]
    wall_blocks_list = {}
    for block_id in block_ids:
        nodes = np.array([i for i in range(Nnodes) if (neighboring_blocks_ids[i, :]==block_id).any()])
        blockG = nx.Graph()
        for node in nodes:
            node_edges = Gcomplete.edges(node)
            for edge in node_edges:
                if not blockG.has_edge(edge[0], edge[1]):
                    blockG.add_edge(edge[0], edge[1])
        
        # prune nodes not belonging to block
        blockG_nodes = np.array(blockG.nodes())
        for node in blockG_nodes:
            if not (nodes==node).any():
                blockG.remove_node(node)

        for node in nodes:
            node_edges = blockG.edges(node)
            
            if len(node_edges)==1:
                print("!!!!! WARNING !!!! BUILT BLOCK HAS OPEN EDGE AT ", node)

        wall_blocks_list.update({block_id: blockG.copy()})

    return wall_blocks_list

def plot_wall_blocks(wall_blocks_list, nodes_coords_complete, with_labels=False, node_size=1, graph_cmap=cm.Dark2, ax=None, imgw=6, imgh=3):
    """
    Plot exploded view of the wall.
    Inputs:
        wall_blocks_list: list of dicts {block_id: block_graph}
        nodes_coords_complete: Nnodes x 2 array
        # plot options
        with_labels: if node labels are to be plot
        node_size: size of node
        graph_cmap: colormap to be used
        ax: specify axis in which to plot
        imgw: image width
        imgh: image height
    """
    font_size = 3
    width = 0.25
    
    # Because positive row indices point down, we flip the y axis
    nodes_coords_complete[:, 1] = -1*nodes_coords_complete[:, 1]
    
    Nblocks = len(wall_blocks_list.keys())
    graph_colors = graph_cmap(np.linspace(0, 1, Nblocks))
    
    wall_xmin = np.amin(nodes_coords_complete[:, 0])
    wall_xmax = np.amax(nodes_coords_complete[:, 0])
    
    wall_ymin = np.amin(nodes_coords_complete[:, 1])
    wall_ymax = np.amax(nodes_coords_complete[:, 1])
    
    wall_xcenter = 0.5*(wall_xmin + wall_xmax)
    wall_ycenter = 0.5*(wall_ymin + wall_ymax)
    wall_center = np.array([wall_xcenter, wall_ycenter])
    
    scx = 0.3
    scy = scx*(wall_xmax - wall_xmin) / (wall_ymax - wall_ymin)
    sc = np.array([scx, scy])
    
    if ax is None:
        plt.figure(dpi=dpi)
        ax = plt.gca()

    c = 0
    for block_id, blockG in wall_blocks_list.items():
        graph_color = graph_colors[c, :][np.newaxis, ...]
        alpha = 0.2
        patch_color = graph_color.flatten()
        patch_color[-1] = alpha
        c = c + 1
        nodes = np.array(blockG.nodes()).astype(np.int32)
        
        diG = get_clockwise_G(blockG, nodes_coords_complete)
        diG_nodes = np.array(diG.nodes()).astype(np.int32)
        block_nodes_coords = nodes_coords_complete[diG_nodes, :]
        
        nodes_xmin = np.amin(nodes_coords_complete[nodes, 0])
        nodes_xmax = np.amax(nodes_coords_complete[nodes, 0])
        nodes_xcenter = 0.5*(nodes_xmin + nodes_xmax)
        
        nodes_ymin = np.amin(nodes_coords_complete[nodes, 1])
        nodes_ymax = np.amax(nodes_coords_complete[nodes, 1])
        
        nodes_ycenter = 0.5*(nodes_ymin + nodes_ymax)
        block_center = np.array([nodes_xcenter, nodes_ycenter])
        
        ann_pos = (nodes_xcenter + (nodes_xcenter-wall_xcenter)*scx, nodes_ycenter + (nodes_ycenter-wall_ycenter)*scy)
        pos = {}
        for node in nodes:
            pos.update({node: nodes_coords_complete[node, :] + (block_center - wall_center)*sc})
        nx.draw_networkx(blockG, pos=pos, ax=ax, with_labels=with_labels, node_size=node_size, node_color=graph_color, edge_color=graph_color, font_size=font_size, font_color='b', alpha=0.95, linewidths=0, width=width)
        ax.annotate(str(int(np.round(block_id))), xy=ann_pos,  xycoords='data', xytext=ann_pos, textcoords='data', horizontalalignment='center', verticalalignment='center', size=4, color=(graph_color[0,0], graph_color[0,1], graph_color[0,2]))
        poly = mpl.patches.Polygon(block_nodes_coords + (block_center - wall_center)*sc, color=patch_color, edgecolor=None)
        ax.add_patch(poly)
        
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.set_aspect('equal')
    plt.axis('off')
    plt.tight_layout()
    plt.gcf().set_size_inches(imgw//2, imgh//2)
    plt.draw()
    plt.pause(0.001)
    
    return None

def get_clockwise_G(blockG, nodes_coords):
    """
    Return a clockwise direct graph.
    Inputs:
        blockG: a graph from a block.
        nodes_coords: Nnodes x 2 array
    Output:
        diG: directed block graph
    """
    block_nodes = np.array(blockG.nodes()).astype(np.int32)

    block_nodes_coords = nodes_coords[block_nodes, :]
   
    ymax = np.amax(block_nodes_coords[:, 1])

    top_nodes = np.where(block_nodes_coords[:, 1]==ymax)[0]
    block_top_nodes = block_nodes[top_nodes]
    
    top_nodes_x = block_nodes_coords[top_nodes, 0]
    
    sorty = np.argsort(top_nodes_x)
    block_top_nodes = block_top_nodes[sorty]
    block_top_nodes = block_top_nodes[:2]

    diG = nx.DiGraph()
    diG.add_edge(block_top_nodes[0], block_top_nodes[1])
    
    end_node = block_top_nodes[0]
    cur_node = block_top_nodes[1]
    circled = False
    can_connect = False
    
    c = 0
    while not circled:
        c = c + 1
        node_edges = blockG.edges(cur_node)
        for edge in node_edges:
            next_node = edge[1]
            if not diG.has_node(next_node):
                diG.add_edge(cur_node, next_node)
                cur_node = next_node
            
            if can_connect and next_node==end_node:
                diG.add_edge(cur_node, end_node)
                circled = True
        
        if len(blockG.nodes())==len(diG.nodes()):
            can_connect = True

        if c > 50:
            # To avoid infinite
            circled = True
    return diG


def get_map_to_relabel(nodes_coords_complete, nodes_coords):
    """
    Map node enumeration from the incomplete to the complete graph.
    Inputs:
        nodes_coords_complete: Nnodes_complete x 2
        nodes_coords: Nnodes x 2
    Output:
        mapping: dict {v_old: v_new}
    """
    Nnode = nodes_coords.shape[0]
    c = Nnode
    
    mapping = {}
    for v_old, node_coord_complete in enumerate(nodes_coords_complete):
        dpos = np.abs(nodes_coords - node_coord_complete)
        dpos = dpos[:, 0] + dpos[:, 1]
        
        match = np.where(dpos==0.)[0]
        if len(match) > 0:
            v_new = match[0]
        else:
            v_new = c
            c = c + 1
        
        mapping.update({v_old: v_new})
        
    return mapping
    
def relabel_Gcomplete(Gcomplete, nodes_coords_complete, nodes_coords):
    """
    Relabel the complete graph nodes to match those of the incomplete graph.
    Inputs:
        Gcomplete: graph of all edges and corners of a sample wall
        nodes_coords_complete: nodes coordinates of the complete graph
        nodes_coords: nodes coordinates of the incomplete graph
    Outputs:
        Gcomplete: graph of all edges and corners of a sample wall
        new_nodes_coords_complete: nodes coordinates of the complete graph
    """

    new_nodes_coords_complete = np.zeros(nodes_coords_complete.shape)
    mapping = get_map_to_relabel(nodes_coords_complete, nodes_coords)
    
    for v_old, v_new in mapping.items():
        new_nodes_coords_complete[v_new, :] = nodes_coords_complete[v_old, :]
    
    Gcomplete = nx.relabel_nodes(Gcomplete, mapping, copy=True)

    return Gcomplete, new_nodes_coords_complete

def add_lintels(target_space, lintel_id, added_lintel_ids=np.array([]), added_lintel_sizes = None, lintel_height=1, lintel_ext=1, lintel_depth=None):
    """
    Add lintels over openings of the target_space
    Inputs:
        target_space: H x W x D array, the target volume
        lintel_id: integer, must be larger than the number of block types
        added_lintel_ids: list of lintel instance ids
        added_lintel_sizes: list of lintel instance sizes
        lintel_height, lintel_ext, lintel_depth: 
    Outputs:
        target_space: H x W x D array
        added_lintel_ids: list of lintel instance ids
        added_lintel_sizes: list of lintel instance sizes
    """

    # We find the position of horizontal openings using convolution on the image
    kernel = [
        [-1],
        [1],
    ]
    
    remove_first_row = False
    if added_lintel_sizes is None:
        remove_first_row = True
        added_lintel_sizes = np.array([0,0,0])

    convo_mode = 'valid'
    for k in range(0, target_space.shape[2], lintel_depth):
        target_wall = target_space[:,:,k]

        padded_target_wall = np.zeros((target_wall.shape[0]+1, target_wall.shape[1]))
        padded_target_wall[:-1, :] = np.abs(target_wall)
        padded_wall_lintels = signal.convolve2d(padded_target_wall, kernel, mode=convo_mode)
        wall_lintels = np.zeros(padded_wall_lintels.shape)
        wall_lintels[np.where(padded_wall_lintels>0)] = 1

        wall_lintels = np.roll(wall_lintels, 1, axis=0)
        Nrows = wall_lintels.shape[0]

        # Extend lintels by lintel_ext
        for ii in range(lintel_ext):
            wall_lintels = wall_lintels + np.roll(wall_lintels, 1, axis=1)
            wall_lintels = wall_lintels + np.roll(wall_lintels, -1, axis=1)
            wall_lintels[np.where(wall_lintels>1)] = 1

        for i, row in enumerate(wall_lintels):
            if (row>0).any():
                prev_col = 0
                for j, col in enumerate(row):
                    if col==1:
                        lintel_width = np.argmin(row[j:]==1)
                        new_lintel_id = get_new_block_id(lintel_id, added_lintel_ids)
                        added_lintel_ids = np.append(added_lintel_ids, new_lintel_id)
                        wall_lintels[i:(i+lintel_height),j:(j+lintel_width)] = new_lintel_id
                        added_lintel_sizes = np.vstack((added_lintel_sizes, [lintel_width, lintel_height, lintel_depth]))

        for kk in range(k, k+lintel_depth):
            target_space[:,:,kk] = target_space[:,:,kk] + wall_lintels

    if remove_first_row:
        added_lintel_sizes = added_lintel_sizes[1:, :]

    return target_space, added_lintel_ids, added_lintel_sizes

def generate_block_vertices(xmin, xmax, ymin, ymax, zmin, zmax):
    """
    Auxiliary function to generate_obj_files.
    """
    block_vertices = [
        [xmax, ymin, zmin],
        [xmax, ymax, zmin],
        [xmax, ymax, zmax],
        [xmax, ymin, zmax],
        [xmin, ymin, zmin],
        [xmin, ymax, zmin],
        [xmin, ymax, zmax],
        [xmin, ymin, zmax],
    ]

    block_normals = [
        [ 1.0,  0.0,  0.0],
        [ 0.0,  1.0,  0.0],
        [ 0.0,  0.0,  1.0],
        [-1.0,  0.0,  0.0],
        [ 0.0, -1.0,  0.0],
        [ 0.0,  0.0, -1.0],
    ]

    return block_vertices, block_normals


def generate_obj_file(filename, target_space, block_sizes, s, lintel_id, cg_offset=np.array([0,0,0]), angle=0.):
    """
    Create .obj file of the target space.
    Inputs:
        filename: output filename
        target_space: H x W x D array
        block_sizes: 
        s: scaling
        lintel_id: id used for lintels
        cg_offset: offset applied to the target space
        angle: rotation applied to the target space
    """

    with_labels = True
    node_size = 1

    Q = np.array([
        [np.cos(angle), -np.sin(angle)],
        [np.sin(angle), np.cos(angle)],
    ])

    Ntarget_rows = target_space.shape[0]
    Ntarget_cols = target_space.shape[1]
    Ntarget_depth = target_space.shape[2]

    xc = Ntarget_cols/2.
    yc = Ntarget_rows/2.
    zc = Ntarget_depth/2.

    ctr_space = np.array([xc, yc, zc])
    for i, coord in enumerate(cg_offset):
        if coord==0.:
            cg_offset[i] = ctr_space[i]

    pos_offset = cg_offset
    added_blocks_ids = []

    with open(filename + ".obj", "a") as f:
        f.write("# " + filename + "\n")
        v_count = 0
        vn_count = 0
        for k in range(Ntarget_depth):
            target_wall = target_space[:, :, k]
            corners_and_edges, corners_mod = get_corners_and_edges(target_wall)
            G, nodes_coords, coords_order = get_wall_graph(corners_and_edges, corners_mod)
            
            Gcomplete, nodes_coords_complete, coords_order_complete = get_complete_wall_graph(corners_and_edges, corners_mod)
            Gcomplete, nodes_coords_complete = relabel_Gcomplete(Gcomplete, nodes_coords_complete, nodes_coords)
            shortestG_list_complete, shortest_pos_list_complete, shortest_pos_list_ordered_complete, nodelist_list_complete, pos_complete, shortest_crack_lengths_complete = get_shortest_paths(Gcomplete, nodes_coords_complete, coords_order_complete, target_wall)
            
            block_ids = np.unique(target_wall)
            neighboring_blocks_ids = get_neighboring_blocks_ids(target_wall, nodes_coords_complete)
            wall_blocks_list = get_wall_blocks(block_ids, neighboring_blocks_ids, Gcomplete)

            for block_id, blockG in wall_blocks_list.items():
                if (added_blocks_ids==block_id).any() or (block_id<0):
                    continue
                else:
                    added_blocks_ids.append(block_id)

                block_type = np.round(block_id).astype(np.int32)
                blockG_nodes = blockG.nodes()
                block_nodes_coords = nodes_coords_complete[blockG_nodes, :]

                nodes_xmin = np.amin(block_nodes_coords[:, 0]).astype(np.int32) + 1
                nodes_xmax = np.amax(block_nodes_coords[:, 0]).astype(np.int32) + 1

                nodes_ymin = np.amin(block_nodes_coords[:, 1]).astype(np.int32)
                nodes_ymax = np.amax(block_nodes_coords[:, 1]).astype(np.int32)

                try:
                    zmin = k*s
                    zmax = zmin + s*block_sizes[block_type, 2]
                except:
                    if np.round(block_id)==lintel_id:
                        zmin=0
                        zmax=s*target_space.shape[2]
                    else:
                        print("Failed to add block")

                node_xc = 0.5*(nodes_xmin+nodes_xmax)
                node_yc = 0.5*(nodes_ymin+nodes_ymax)


                if zmax > s*target_space.shape[2]:
                    zmax = s*target_space.shape[2]
                node_zc = 0.5*(zmin+zmax)
                ctr_node = np.array([node_xc, node_yc, node_zc])
                
                lamb = 0.
                dpos = lamb*(ctr_node - ctr_space)

                block_vertices, block_normals = generate_block_vertices(nodes_xmin + dpos[0] - ctr_space[0] + pos_offset[0], nodes_xmax + dpos[0] - ctr_space[0] + pos_offset[0], nodes_ymin + dpos[1] - ctr_space[1] + pos_offset[1], nodes_ymax + dpos[1] - ctr_space[1] + pos_offset[1], zmin + dpos[2] - s*ctr_space[2] + pos_offset[2], zmax + dpos[2] - s*ctr_space[2] + pos_offset[2])

                block_vertices = np.array(block_vertices)

                f.write("\no cube\n")

                # rotate the coordinates
                for vertice in block_vertices:
                    xz = np.array([[vertice[0]-pos_offset[0]], [vertice[2]-pos_offset[2]]])
                    xz = Q @ xz
                    vertice[0] = xz[0,0] + pos_offset[0]
                    vertice[2] = xz[1,0] + pos_offset[2]

                for vertice in block_vertices:
                    xcoord = str(np.round(10*vertice[0])/10)
                    ycoord = str(np.round(10*vertice[1])/10)
                    zcoord = str(np.round(10*vertice[2])/10)
                    f.write("v " + xcoord + " " + ycoord + " " + zcoord + "\n")
                    

                f.write("\n")
                for normal in block_normals:
                    xcoord = str(np.round(10*vertice[0])/10)
                    ycoord = str(np.round(10*vertice[1])/10)
                    zcoord = str(np.round(10*vertice[2])/10)
                    f.write("vn " + xcoord + " " + ycoord + " " + zcoord + "\n")

                f.write("\n")
                    
                v1 = str(1 + v_count) 
                v2 = str(2 + v_count)
                v3 = str(3 + v_count)
                v4 = str(4 + v_count)
                v5 = str(5 + v_count)
                v6 = str(6 + v_count)
                v7 = str(7 + v_count)
                v8 = str(8 + v_count)

                vn1 = str(1 + vn_count)
                vn2 = str(2 + vn_count)
                vn3 = str(3 + vn_count)
                vn4 = str(4 + vn_count)
                vn5 = str(5 + vn_count)
                vn6 = str(6 + vn_count)

                f.write("f "+v1+"//"+vn1+" "+v2+"//"+vn1+" "+v3+"//"+vn1+" "+v4+"//"+vn1+ "\n")
                f.write("f "+v2+"//"+vn2+" "+v6+"//"+vn2+" "+v7+"//"+vn2+" "+v3+"//"+vn2+ "\n")
                f.write("f "+v4+"//"+vn3+" "+v3+"//"+vn3+" "+v7+"//"+vn3+" "+v8+"//"+vn3+ "\n")
                f.write("f "+v5+"//"+vn4+" "+v6+"//"+vn4+" "+v7+"//"+vn4+" "+v8+"//"+vn4+ "\n")
                f.write("f "+v5+"//"+vn5+" "+v1+"//"+vn5+" "+v4+"//"+vn5+" "+v8+"//"+vn5+ "\n")
                f.write("f "+v1+"//"+vn6+" "+v2+"//"+vn6+" "+v6+"//"+vn6+" "+v5+"//"+vn6+ "\n")

                v_count = v_count + 8
                vn_count = vn_count + 6

        f.close()

    return None

def global_texture_quality(target_space, fill_opts):
    """
    Get the shortest crack path over walls of a target_space.
    Inputs:
        target_space: H x W x D array
        fill_opts: dict of filling options
    Outputs:
        shortest_crack_length: shortest top to bottom path across edges
    """

    with_labels = False # node labels
    node_size = 1
    font_size = 3
    width = 0.25
    graph_color = 'lightblue'

    header_depth = fill_opts["header_depth"]
    target_path_length = fill_opts["target_path_length"]
    minimum_path_height = fill_opts["minimum_path_height"]
    Ntarget_rows, Ntarget_cols, Ntarget_depth = target_space.shape

    shortest_crack_lengths=[]
    for k in range(0, Ntarget_depth, header_depth):
        target_wall = target_space[:, :, k]
        corners_and_edges, corners_mod = get_corners_and_edges(target_wall)
        G, nodes_coords, coords_order = get_wall_graph(corners_and_edges, corners_mod)

        Gcomplete, nodes_coords_complete, coords_order_complete = get_complete_wall_graph(corners_and_edges, corners_mod)
        Gcomplete, nodes_coords_complete = relabel_Gcomplete(Gcomplete, nodes_coords_complete, nodes_coords)
        shortestG_list_complete, shortest_pos_list_complete, shortest_pos_list_ordered_complete, nodelist_list_complete, pos_complete, shortest_crack_lengths_complete = get_shortest_paths(Gcomplete, nodes_coords_complete, coords_order_complete, target_wall)
        
        col_range = np.array([np.amin(nodes_coords_complete[:,0]), np.amax(nodes_coords_complete[:,0])])
        row_range = np.array([np.amin(nodes_coords_complete[:,1]), np.amax(nodes_coords_complete[:,1])])

        Gtrimmed = Gcomplete.copy()
        Gborders = Gcomplete.copy() # windows and doors
        pos_trimmed = pos_complete.copy()
        pos_borders = pos_complete.copy()
        block_ids = np.unique(target_wall)
        neighboring_blocks_ids = get_neighboring_blocks_ids(target_wall, nodes_coords_complete)

        for node, neighboring_blocks_id in enumerate(neighboring_blocks_ids):
            if (neighboring_blocks_id>=-1).all():
                Gborders.remove_node(node)
                pos_borders.pop(node)

        for node, node_coord in enumerate(nodes_coords_complete):
            node_col = node_coord[0].astype(np.int32)
            node_row = node_coord[1].astype(np.int32)

            if (col_range==node_col).any():
                Gtrimmed.remove_node(node)
                pos_trimmed.pop(node)

            if (row_range==node_row).any():
                if Gtrimmed.has_node(node):
                    node_edges = [edge for edge in Gtrimmed.edges(node)]

                    for node_edge in node_edges:
                        y0 = nodes_coords_complete[node_edge[0], 1]
                        y1 = nodes_coords_complete[node_edge[1], 1]
                        if (y0==y1):
                            Gtrimmed.remove_edge(node_edge[0], node_edge[1])
            
            try:
                left_block = target_wall[node_row - 1, node_col-1] 
                right_block = target_wall[node_row - 1, node_col+1] 
                if (left_block<0) or (right_block<0):
                    Gtrimmed.remove_node(node)
                    pos_trimmed.pop(node)
            except:
                continue

        # Remove nodes without connections
        remaining_nodes = [node for node in Gtrimmed.nodes()]
        for node in remaining_nodes:
            if len(Gtrimmed.edges(node))==0:
                Gtrimmed.remove_node(node)
        
        borders_nodes = [node for node in Gborders.nodes()]

        # Find top and bottom nodes
        remaining_nodes = [node for node in Gtrimmed.nodes()]

        top_nodes = np.intersect1d(borders_nodes, remaining_nodes)

        bottom_nodes = np.array([])
        for node in remaining_nodes:
            node_row = nodes_coords_complete[node, 1].astype(np.int32)
            if (row_range[1]==node_row):
                top_nodes = np.append(top_nodes, node)
            elif (row_range[0]==node_row):
                bottom_nodes = np.append(bottom_nodes, node)

        shortest_paths = []
        for bottom_node in bottom_nodes:
            for top_node in top_nodes:
                if Gtrimmed.has_node(bottom_node) and Gtrimmed.has_node(top_node):
                    try:
                        path_length = nx.shortest_path_length(Gtrimmed, bottom_node, top_node, weight="weight")
                        bottom_node_y = nodes_coords_complete[bottom_node.astype(np.int32), 1]
                        top_node_y = nodes_coords_complete[top_node.astype(np.int32), 1]
                        H = top_node_y - bottom_node_y
                        if H > minimum_path_height:
                            shortest_paths.append([bottom_node, top_node, path_length/H])
                    except:
                        print("Error in computing path length -- ")
                        print('Bottom:', bottom_node, 'Top:', top_node)
                        continue

        shortest_paths = np.array(shortest_paths)
        min_shortest_paths = np.amin(shortest_paths[:,2])
        
        shortest_crack_lengths.append(min_shortest_paths)
        min_idxs = np.where(shortest_paths[:,2]==min_shortest_paths)[0]

        shortestG_list = []
        shortest_pos_list = []
        shortest_pos_list_ordered = []
        nodelist_list = []

        shortestGcolors = []
        shortest_rel_difs = []
        for min_idx in range(len(shortest_paths)):
            path_length = shortest_paths[min_idx, 2]
            rel_dif = (target_path_length - path_length) + 1

            if rel_dif < 0:
                rel_dif = 0

            Gcolor = cm.jet(rel_dif)

            if rel_dif>1:
                rel_dif = 3
                Gcolor = (0., 0., 0., 1)
                
            shortestGcolors.append(Gcolor)
            shortest_rel_difs.append(rel_dif)

            nodelist = np.array(nx.shortest_path(Gtrimmed, shortest_paths[min_idx, 0], shortest_paths[min_idx, 1], weight="weight")).astype(np.int32)
            Nshort = len(nodelist)
            shortestGtrimmed = nx.Graph()

            for i in range(0, Nshort-1):
                edge_weight = Gtrimmed.get_edge_data(nodelist[i], nodelist[i+1])["weight"]
                shortestGtrimmed.add_edge(nodelist[i], nodelist[i+1], weight=edge_weight)

            shortest_coords = nodes_coords_complete[nodelist, :].copy()
            shortest_pos = {}
            for k, coord in enumerate(shortest_coords):
                shortest_pos.update({nodelist[k]: coord + 2*(np.random.rand()-0.5)})

            shortestG_list.append(shortestGtrimmed)
            shortest_pos_list.append(shortest_pos)
            shortest_pos_list_ordered.append(shortest_coords)
            nodelist_list.append(nodelist)

    shortest_crack_lengths = np.array(shortest_crack_lengths)
    shortest_crack_length = np.amin(shortest_crack_lengths)

    return shortest_crack_length

def generate_parts_list(target_space, block_sizes, s, lintel_id, lintel_ids, lintel_sizes, cg_offset=np.array([0,0,0]), angle=0.):
    """
    Generates data for reconstruction in Abaqus.
    """
    with_labels = True
    node_size = 1

    Q = np.array([
        [np.cos(angle), -np.sin(angle)],
        [np.sin(angle), np.cos(angle)],
    ])

    # for k in walls_to_fill:
    Ntarget_rows = target_space.shape[0]
    Ntarget_cols = target_space.shape[1]
    Ntarget_depth = target_space.shape[2]

    xc = Ntarget_cols/2.
    yc = Ntarget_rows/2.
    zc = Ntarget_depth/2.

    target_space_range = np.array([
        [-xc/s, xc/s],
        [-yc/s, yc/s],
        [-zc, zc],
    ])

    ctr_space = np.array([xc, yc, zc])

    for i, coord in enumerate(cg_offset):
        if coord==0.:
            cg_offset[i] = ctr_space[i]

    pos_offset = cg_offset # - ctr_space
    # print("post:", cg_offset)
    # print(pos_offset)

    # grid_size = np.ceil(np.sqrt(Ntarget_depth)).astype(np.int32)
    # axs = axs.flatten()

    added_blocks_ids = []
    added_blocks_geom = []
    added_blocks_translate = []

    for k in range(Ntarget_depth):
        # fig, ax = plt.subplots(1, 1, dpi=dpi)
        target_wall = target_space[:, :, k]
        corners_and_edges, corners_mod = get_corners_and_edges(target_wall)
        G, nodes_coords, coords_order = get_wall_graph(corners_and_edges, corners_mod)
        
        Gcomplete, nodes_coords_complete, coords_order_complete = get_complete_wall_graph(corners_and_edges, corners_mod)
        Gcomplete, nodes_coords_complete = relabel_Gcomplete(Gcomplete, nodes_coords_complete, nodes_coords)
        shortestG_list_complete, shortest_pos_list_complete, shortest_pos_list_ordered_complete, nodelist_list_complete, pos_complete, shortest_crack_lengths_complete = get_shortest_paths(Gcomplete, nodes_coords_complete, coords_order_complete, target_wall)
        # plot_wall(Gcomplete, pos_complete, shortestG_list_complete, shortest_pos_list_complete, target_wall, round_values=True, with_labels=with_labels, node_size=node_size, graph_color='k')
        
        block_ids = np.unique(target_wall)
        neighboring_blocks_ids = get_neighboring_blocks_ids(target_wall, nodes_coords_complete)
        wall_blocks_list = get_wall_blocks(block_ids, neighboring_blocks_ids, Gcomplete)
        # plot_wall_blocks(wall_blocks_list, nodes_coords_complete, node_size=0.5, with_labels=False, ax=ax, imgw=8, imgh=8)

        for block_id, blockG in wall_blocks_list.items():
            if (added_blocks_ids==block_id).any() or (block_id<0):
                continue
            else:
                added_blocks_ids.append(block_id)

            block_type = np.round(block_id).astype(np.int32)
            blockG_nodes = blockG.nodes()
            block_nodes_coords = nodes_coords_complete[blockG_nodes, :]

            nodes_xmin = np.amin(block_nodes_coords[:, 0]).astype(np.int32) + 1
            nodes_xmax = np.amax(block_nodes_coords[:, 0]).astype(np.int32) + 1

            nodes_ymin = np.amin(block_nodes_coords[:, 1]).astype(np.int32)
            nodes_ymax = np.amax(block_nodes_coords[:, 1]).astype(np.int32)

            zmin = k*s
            try:
                zmax = zmin + s*block_sizes[block_type, 2]
            except:
                if np.round(block_id)==lintel_id:
                    cur_lintel = np.where(lintel_ids==block_id)[0][0]
                    lintel_size = lintel_sizes[cur_lintel, :]
                    zmax=s*lintel_size[2]
                else:
                    print("Failed to add block")

            node_xc = 0.5*(nodes_xmin+nodes_xmax)
            node_yc = 0.5*(nodes_ymin+nodes_ymax)
            if zmax > s*target_space.shape[2]:
                zmax = s*target_space.shape[2]

            # node_zc = 0.5*(zmin+zmax)
            node_zc = zmin
            ctr_node = np.array([node_xc, node_yc, node_zc])

            block_geom = [nodes_xmin-node_xc, nodes_ymax-node_yc, nodes_xmax-node_xc, nodes_ymax-node_yc, zmax-zmin]

            translate_vector = ctr_node - ctr_space
            added_blocks_translate.append(translate_vector)
            added_blocks_geom.append(block_geom)

    added_blocks_translate = np.array(added_blocks_translate)/s
    added_blocks_ids = np.array(added_blocks_ids)
    added_blocks_geom = np.array(added_blocks_geom)
    return added_blocks_ids, added_blocks_translate, added_blocks_geom, target_space_range

def merge_blocks(target_space, block_ids, fill_blocks_list, block_sizes, fill_opts):
    """
    Attempts to merge blocks in block_list with neighboring blocks.
    Inputs:
        target_space: H x W x D array
        block_ids: array of block instance ids
        blocks_list: list of block types that can be used for filling gaps
        block_sizes: Nblocks x 3 array
        fill_opts: filling options
    Outputs:
        target_space: H x W x D array
        block_ids: array of block instance ids
    """

    s = fill_opts["scale"]
    header_height = fill_opts["header_height"]
    header_depth = fill_opts["header_depth"]
    header_pad = fill_opts["header_pad"]
    header_init_max = fill_opts["header_init_max"]
    header_init_sizes = fill_opts["header_init_sizes"]
    col_pad = fill_opts["col_pad"]
    row_pad = fill_opts["row_pad"]
    depth_pad = fill_opts["depth_pad"]
    lintel_id = fill_opts["lintel_id"]

    Ntarget_rows, Ntarget_cols, Ntarget_depth = target_space.shape
    block_ids = np.unique(target_space)

    for i in range(0, Ntarget_rows, s*header_height):
        for j in range(0, Ntarget_cols, s):
            for k in range(0, Ntarget_depth, header_depth):
                cur_block_id = target_space[i,j,k]
                cur_block_type = np.round(cur_block_id).astype(np.int32)

                if (fill_blocks_list==cur_block_type).any():
                    cur_block_width = block_sizes[cur_block_type, 0]
                    cur_block_height = block_sizes[cur_block_type, 1]
                    cur_block_depth = block_sizes[cur_block_type, 2]

                    prev_block_id = target_space[i, j-s, k]
                    prev_block_type = np.round(prev_block_id).astype(np.int32)

                    if prev_block_id!=cur_block_id and (cur_block_id>0) and (prev_block_id>0):
                        if prev_block_type!=lintel_id:
                            prev_block_width = block_sizes[prev_block_type, 0]
                            prev_block_height = block_sizes[prev_block_type, 1]
                            prev_block_depth = block_sizes[prev_block_type, 2]

                            if (cur_block_height==prev_block_height) and (cur_block_depth==prev_block_depth):
                                merged_block_width = cur_block_width + prev_block_width
                                possible_fill_blocks = np.where((block_sizes[:,0]==merged_block_width)&(block_sizes[:,1]==cur_block_height)&(block_sizes[:,2]==cur_block_depth))[0]

                                if len(possible_fill_blocks>0):
                                    blockN = possible_fill_blocks[0]
                                    new_block_id = get_new_block_id(blockN, block_ids)
                                    block_ids = np.append(block_ids, new_block_id)
                                    target_space[np.where(target_space==cur_block_id)] = new_block_id
                                    target_space[np.where(target_space==prev_block_id)] = new_block_id

    return target_space, block_ids

def main():
    # Based on Guimaraes castle
    sample_blocks = np.array([
    [6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 42, 42, 42, 42, 8, 8, 8, 8, 8, 8, 8, 8, 42, 42, 42, 42],
    [6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 42, 42, 42, 42, 8, 8, 8, 8, 8, 8, 8, 8, 42, 42, 42, 42],
    [6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 42, 42, 42, 42, 8, 8, 8, 8, 8, 8, 8, 8, 42, 42, 42, 42],
    [6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 42, 42, 42, 42, 8, 8, 8, 8, 8, 8, 8, 8, 42, 42, 42, 42],
    [32, 32, 32, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 8, 8, 8, 8, 8, 8, 8, 8],
    [32, 32, 32, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 8, 8, 8, 8, 8, 8, 8, 8],
    [32, 32, 32, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 8, 8, 8, 8, 8, 8, 8, 8],
    [32, 32, 32, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 8, 8, 8, 8, 8, 8, 8, 8],
    [8, 8, 8, 8, 8, 8, 8, 8, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5],
    [8, 8, 8, 8, 8, 8, 8, 8, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5],
    [8, 8, 8, 8, 8, 8, 8, 8, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5],
    [8, 8, 8, 8, 8, 8, 8, 8, 4, 4, 4, 4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5],
    [4, 4, 4, 4, 32, 32, 32, 42, 42, 42, 42, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
    [4, 4, 4, 4, 32, 32, 32, 42, 42, 42, 42, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
    [4, 4, 4, 4, 32, 32, 32, 42, 42, 42, 42, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
    [4, 4, 4, 4, 32, 32, 32, 42, 42, 42, 42, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 32, 32, 32, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5],
    [8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 32, 32, 32, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5],
    [8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 32, 32, 32, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5],
    [8, 8, 8, 8, 8, 8, 8, 8, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 32, 32, 32, 5, 5, 5, 5, 5, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5],
    [7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [7, 7, 7, 7, 7, 7, 7, 32, 32, 32, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, 5, 5, 5, 5, 42, 42, 42, 42, 4, 4, 4, 4, 5, 5, 5, 5, 5, 9, 9, 9, 9, 9, 9, 9, 9, 9, 6, 6, 6, 6, 6, 6],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, 5, 5, 5, 5, 42, 42, 42, 42, 4, 4, 4, 4, 5, 5, 5, 5, 5, 9, 9, 9, 9, 9, 9, 9, 9, 9, 6, 6, 6, 6, 6, 6],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, 5, 5, 5, 5, 42, 42, 42, 42, 4, 4, 4, 4, 5, 5, 5, 5, 5, 9, 9, 9, 9, 9, 9, 9, 9, 9, 6, 6, 6, 6, 6, 6],
    [5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, 5, 5, 5, 5, 42, 42, 42, 42, 4, 4, 4, 4, 5, 5, 5, 5, 5, 9, 9, 9, 9, 9, 9, 9, 9, 9, 6, 6, 6, 6, 6, 6],
    ])

    # Original
    block_sizes = np.array([ #[width, height, depth]
        [3, 4, 8], #32, 0
        [4, 4, 4], #4, 1
        [4, 4, 8], #42, 2
        [5, 4, 4], #5, 3
        [6, 4, 4], #6, 4
        [7, 4, 4], #7, 5
        [8, 4, 4], #8, 6
        [9, 4, 4], #9, 7
        [10, 4, 4], #10, 8
        [11, 4, 4], #11, 9
        [2, 4, 4],# 10
        [1, 4, 4], # 11
        [3, 4, 4], # 12
    ])

    # # No through
    # block_sizes = np.array([ #[width, height, depth]
    #     [3, 4, 4], #32, 0
    #     [4, 4, 4], #4, 1
    #     [4, 4, 4], #42, 2
    #     [5, 4, 4], #5, 3
    #     [6, 4, 4], #6, 4
    #     [7, 4, 4], #7, 5
    #     [8, 4, 4], #8, 6
    #     [9, 4, 4], #9, 7
    #     [10, 4, 4], #10, 8
    #     [11, 4, 4], #11, 9
    #     [2, 4, 4],# 10
    #     [1, 4, 4], # 11
    #     [3, 4, 4], # 12
    # ])

    # # All through
    # block_sizes = np.array([ #[width, height, depth]
    #     [3, 4, 8], #32, 0
    #     [4, 4, 8], #4, 1
    #     [4, 4, 8], #42, 2
    #     [5, 4, 8], #5, 3
    #     [6, 4, 8], #6, 4
    #     [7, 4, 8], #7, 5
    #     [8, 4, 8], #8, 6
    #     [9, 4, 8], #9, 7
    #     [10, 4, 8], #10, 8
    #     [11, 4, 8], #11, 9
    #     [2, 4, 8],# 10
    #     [1, 4, 8], # 11
    #     [3, 4, 8], # 12
    # ])

    Nsample_rows, Nsample_cols = sample_blocks.shape
    # The id of the blocks in sample_blocks should match the order in the block_sizes list, in our example they do not so we will remap it
    block_id_remap = [32, 4, 42, 5, 6, 7, 8, 9, 10, 11, 12]
    aux_sample_blocks = sample_blocks.copy()
    for j, block_id in enumerate(block_id_remap):
        sample_blocks[np.where(aux_sample_blocks==block_id)] = j

    # lintel info
    lintel_id = 15
    lintel_height = 4
    lintel_ext = 3 # how many units beyond the openning it should extend
    lintel_depth = 8

    s = 5 
    # options for the filling algorithm
    fill_opts = {
        "scale": s, # Scaling factor along the wall direction, this will be undone at the later stage
        "header_height": 4, # Print head info
        "header_depth": 4,
        "header_pad": 1,
        "header_init_max": 0, # It is possible to initialize the print head at a random position, but here we always start at the topmost leftmost position
        "header_init_sizes": [],
        "col_pad": s*15,
        "row_pad": s*4,
        "depth_pad": 4,
        "lintel_id": lintel_id,
        "target_path_length": 0,
        "minimum_path_height": 0,
    }

    # Each block type has a unique, integer base id (e.g., 1), but each instance of that block type has an unique id that is a float (e.g., 1.01, 1.11, etc.)
    # This function introduces noise so that adjacent blocks of the same type can be distinguished in sample_blocks
    noisy_sample_blocks = noisify_sample(sample_blocks, block_sizes)
    block_sizes = s*block_sizes
    block_sizes[:, 2] = block_sizes[:, 2]/s # Do not scale depth

    # Count number of each element present
    Nblock_types = block_sizes.shape[0]

    # Compute the probability of observing a block based on area ratios
    block_probs = probs(sample_blocks, block_sizes)

     # Fill guimaraes facade volume
    facade = np.ones((72,136,8)) # Target volume space (x,y,z) with z the depth
    facade = -1*facade # Fillable volume is set as -1, ids < -1 are considered opennings

    # Openings
    doors = np.zeros(facade.shape) # Volume storing only doors
    windows = np.zeros(facade.shape) # Volume storing only windows

    # Doors: -2
    # Specify the doors in the facade
    facade[0:20, 17:28,:]=-2
    facade[0:20, 52:66,:]=-2
    facade[0:20, 86:101,:]=-2
    # and in the doors volume
    doors[np.where(facade==-2)] = -2

    # Adds lintels to the doors, and generates the updated door volume, the list of associated lintel ids, and their sizes in the same format as block_sizes
    doors, door_lintel_ids, door_lintel_sizes = add_lintels(doors, lintel_id=lintel_id, added_lintel_sizes = None, lintel_height=lintel_height, lintel_ext=lintel_ext, lintel_depth=lintel_depth)

    # Scaling is not strictly necessary but it was used for some auxiliary algorithms, so it is kept here since our models were generated with this option
    upscaled_doors = upscale_space(doors, sr=s, sc=s, sd=1)
    
    # Windows -3
    facade[8:16, 113:120,:]=-3
    facade[40:64, 9:23,:]=-3
    facade[40:64, 32:47,:]=-3
    facade[36:56, 114:127,:]=-3
    windows[np.where(facade==-3)] = -3
    windows, lintel_ids, lintel_sizes = add_lintels(windows, added_lintel_ids=door_lintel_ids, lintel_id=lintel_id, added_lintel_sizes=door_lintel_sizes, lintel_height=lintel_height, lintel_ext=lintel_ext, lintel_depth=lintel_depth)
    upscaled_windows = upscale_space(windows, sr=s, sc=s, sd=1)

    upscaled_facade = upscale(facade[:,:,0], s, s) # Pick one leaf just to retrieve upscaled sizes

    uvs = s # Unit voxel size
    Ntarget_rows, Ntarget_cols = upscaled_facade.shape
    Ntarget_depth = facade.shape[2]
    longest_width = np.amax(block_sizes[:,0])

    target_wall = upscale(noisy_sample_blocks, s, s)
    ## As a proxy for quality, we use the shortest path from top to bottom of a wall, to do this we form a graph of the block's edges and cornders, and compute the shortest graph path form nodes of the top of the wall to the bottom of the wall.
    corners_and_edges, corners = get_corners_and_edges(target_wall)

    # Returns a graph G of the edges and corners of the blocks within the wall
    G, nodes_coords, coords_order = get_wall_graph(corners_and_edges, corners)

    # Finds the shortest paths from the top to the bottom of the wall
    shortestG_list, shortest_pos_list, shortest_pos_list_ordered, nodelist_list, pos, sample_shortest_crack_lengths = get_shortest_paths(G, nodes_coords, coords_order, target_wall)

    check_area_size = target_wall.shape # Need to better explain this variable
    target_path_length = sample_shortest_crack_lengths
    fill_opts["target_path_length"] = target_path_length[0]
    fill_opts["minimum_path_height"] = target_wall.shape[0]

    Gcomplete, nodes_coords_complete, coords_order_complete = get_complete_wall_graph(corners_and_edges, corners)
    Gcomplete, nodes_coords_complete = relabel_Gcomplete(Gcomplete, nodes_coords_complete, nodes_coords)
    shortestG_list_complete, shortest_pos_list_complete, shortest_pos_list_ordered_complete, nodelist_list_complete, pos_complete, sample_shortest_crack_lengths_complete = get_shortest_paths(Gcomplete, nodes_coords_complete, coords_order_complete, target_wall)

    plot_wall(G, pos, shortestG_list, shortest_pos_list, target_wall, round_values=True, with_labels=True, node_size=1)
    plot_wall(Gcomplete, pos_complete, shortestG_list_complete, shortest_pos_list_complete, target_wall, round_values=True, with_labels=True, node_size=1, graph_color='k')

    block_ids = np.unique(target_wall)
    neighboring_blocks_ids = get_neighboring_blocks_ids(target_wall, nodes_coords_complete)
    wall_blocks_list = get_wall_blocks(block_ids, neighboring_blocks_ids, Gcomplete)

    plot_wall_blocks(wall_blocks_list, nodes_coords_complete, node_size=0.)
    plt.show()

    # Here we generate several samples of the filling, and pick those with longest 'crack paths'
    Nsamples = 10
    plot_every = 1
    path_tol = 0.1

    wall_shortest_crack_lengths = {
        0: np.array([]),
        1: np.array([])
    }
    verbose = True

    target_size = (Ntarget_rows, Ntarget_cols, Ntarget_depth)
    sample_block_probs = probs(sample_blocks, block_sizes)

    walls_block_probs = {}
    for i in range(Ntarget_depth):
        walls_block_probs.update({i: sample_block_probs})

    upscaled_doors_and_windows = upscaled_doors + upscaled_windows
    upscaled_lintels = upscaled_doors_and_windows.copy()
    upscaled_lintels[np.where(upscaled_lintels<lintel_id)] = -1.
    upscaled_doors_and_windows[np.where(upscaled_doors_and_windows==0)] = -1.
    
    # We use a proxy for quality based on the shortest path for selection of the textures
    global_shortest_crack_lengths=[]
    best_global_data = {
        "target_space": None,
        "lintel_id": None,
        "lintel_ids": None,
        "lintel_sizes": None,
    }

    best_fill_shortest_crack_length = 0
    for pp in range(Nsamples):
        padded_target_space = create_target_space(target_size, fill_opts)
        padded_target_space = 2*padded_target_space

        for k in range(Ntarget_depth):
            upscaled_leave = upscaled_doors_and_windows[:,:,k]
            padded_target_space = prefil_padded_target_space(upscaled_leave, padded_target_space, k, fill_opts)

        padded_target_space, block_ids = fill_verticals(padded_target_space, sample_blocks, block_sizes, walls_block_probs, fill_opts)
        padded_target_space = fill_space(padded_target_space, sample_blocks, block_sizes, walls_block_probs, fill_opts, block_ids=block_ids, unit_blockN=-1)
        target_space = remove_target_pad(padded_target_space, fill_opts)
        shortest_crack_length = global_texture_quality(target_space, fill_opts)

        if shortest_crack_length > best_fill_shortest_crack_length:
            best_fill_shortest_crack_length = shortest_crack_length
            best_global_data["target_space"] = target_space.copy()
            best_global_data["lintel_id"] = lintel_id
            best_global_data["lintel_ids"] = lintel_ids.copy()
            best_global_data["lintel_sizes"] = lintel_sizes.copy()

        global_shortest_crack_lengths.append(shortest_crack_length)
        print(global_shortest_crack_lengths)
    
    target_space = best_global_data["target_space"]
    lintel_id = best_global_data["lintel_id"]
    lintel_ids = best_global_data["lintel_ids"]
    lintel_sizes = best_global_data["lintel_sizes"]

    # Generates information for Abaqus reconstruction
    added_blocks_ids, added_blocks_translate, added_blocks_geom, target_space_range = generate_parts_list(target_space, block_sizes, s, lintel_id=lintel_id, lintel_ids=lintel_ids, lintel_sizes=lintel_sizes, cg_offset=np.array([0,0,0]), angle=0.)
    
    # Undo scaling
    block_sizes[:,0] = block_sizes[:,0]/s 
    block_sizes[:,1] = block_sizes[:,1]/s

    # Saves data for reconstruction on Abaqus
    blocklist_filename = 'guimaraes_blocklist_test'
    np.savez(blocklist_filename, block_sizes=block_sizes, blocks_ids=added_blocks_ids, blocks_translate=added_blocks_translate, lintel_ids=lintel_ids, lintel_sizes=lintel_sizes, blocks_geom=added_blocks_geom, target_space_range=target_space_range)

    print("DONE")

if __name__=="__main__":
    main()

